(*========================================================================*)
(*                                                                        *)
(*                 Compositional CompcertTSO                              *)
(*                                                                        *)
(*                 Chung-Kil Hur, MPI-SWS                                 *)
(*                 Viktor Vafeiadis, MPI-SWS                              *)
(*                 Derek Dreyer, MPI-SWS                                  *)
(*                                                                        *)
(*  Copyright 2010                                                        *)
(*                                                                        *)
(*  All rights reserved.  This file is distributed under the terms of     *)
(*  the INRIA Non-Commercial License Agreement.                           *)
(*                                                                        *)
(*========================================================================*)

Require Export Syntax.

Require Import Coqlib.
Require Import Libtactics.

(*************************
         Tactics
 *************************)

Notation " 'exists' a b c d e , p" := 
 (ex (fun a => (ex (fun b => (ex (fun c => (ex (fun d => (ex (fun e => p))))))))))
 (at level 200, a ident, b ident, c ident, d ident, e ident, right associativity) 
 : type_scope.

Notation " 'exists' a b c d e f , p" := 
 (ex (fun a => (ex (fun b => (ex (fun c => (ex (fun d => (ex (fun e => 
   (ex (fun f => p))))))))))))
 (at level 200, a ident, b ident, c ident, d ident, e ident, f ident, right associativity) 
 : type_scope.

Notation " 'exists' a b c d e f g , p" := 
 (ex (fun a => (ex (fun b => (ex (fun c => (ex (fun d => (ex (fun e => 
   (ex (fun f => (ex (fun g => p))))))))))))))
 (at level 200, a ident, b ident, c ident, d ident, e ident, f ident, g ident,
   right associativity) 
 : type_scope.

Notation " 'exists' a b c d e f g h , p" := 
 (ex (fun a => (ex (fun b => (ex (fun c => (ex (fun d => (ex (fun e => 
   (ex (fun f => (ex (fun g => (ex (fun h => p))))))))))))))))
 (at level 200, a ident, b ident, c ident, d ident, e ident, f ident, g ident,
   h ident, right associativity) 
 : type_scope.

Notation " 'exists' a b c d e f g h i , p" := 
 (ex (fun a => (ex (fun b => (ex (fun c => (ex (fun d => (ex (fun e => 
   (ex (fun f => (ex (fun g => (ex (fun h => (ex (fun i => p))))))))))))))))))
 (at level 200, a ident, b ident, c ident, d ident, e ident, f ident, g ident,
   h ident, i ident, right associativity) 
 : type_scope.

Notation " 'exists' a b c d e f g h i j , p" := 
 (ex (fun a => (ex (fun b => (ex (fun c => (ex (fun d => (ex (fun e => 
   (ex (fun f => (ex (fun g => (ex (fun h => (ex (fun i => 
     (ex (fun j => p))))))))))))))))))))
 (at level 200, a ident, b ident, c ident, d ident, e ident, f ident, g ident,
   h ident, i ident, j ident, right associativity) 
 : type_scope.

Tactic Notation "exists" constr(a) constr(b) constr(c) constr(d) constr(e) := 
  exists a; exists b ; exists c ; exists d; exists e.
 
Tactic Notation "exists" constr(a) constr(b) constr(c) constr(d) constr(e) constr(f):= 
  exists a; exists b ; exists c ; exists d; exists e; exists f.

Tactic Notation "exists" constr(a) constr(b) constr(c) constr(d) constr(e) constr(f)
  constr(g) := 
  exists a; exists b ; exists c ; exists d; exists e; exists f; exists g.

Tactic Notation "exists" constr(a) constr(b) constr(c) constr(d) constr(e) constr(f)
  constr(g) constr(h) := 
  exists a; exists b ; exists c ; exists d; exists e; exists f; exists g; exists h.

Tactic Notation "exists" constr(a) constr(b) constr(c) constr(d) constr(e) constr(f)
  constr(g) constr(h) constr(i) := 
  exists a; exists b ; exists c ; exists d; exists e; exists f; exists g; exists h;
  exists i.

Tactic Notation "exists" constr(a) constr(b) constr(c) constr(d) constr(e) constr(f)
  constr(g) constr(h) constr(i) constr(j) := 
  exists a; exists b ; exists c ; exists d; exists e; exists f; exists g; exists h;
  exists i; exists j.

Notation "p <1= q" := 
  (forall a (PR: p a : Prop), q a : Prop) 
  (at level 50, no associativity).

Notation "p <2= q" := 
  (forall a b (PR: p a b : Prop), q a b : Prop) 
  (at level 50, no associativity).

Notation "p <3= q" := 
  (forall a b c (PR: p a b c : Prop), q a b c : Prop) 
  (at level 50, no associativity).

Notation "p <4= q" := 
  (forall a b c d (PR: p a b c d : Prop), q a b c d : Prop) 
  (at level 50, no associativity).

Notation "p <5= q" := 
  (forall a b c d e (PR: p a b c d e : Prop), q a b c d e : Prop) 
  (at level 50, no associativity).

Notation "p <6= q" := 
  (forall a b c d e f (PR: p a b c d e f : Prop), q a b c d e f : Prop) 
  (at level 50, no associativity).

Notation "p <7= q" := 
  (forall a b c d e f g (PR: p a b c d e f g : Prop), q a b c d e f g : Prop) 
  (at level 50, no associativity).

Notation "p <8= q" := 
  (forall a b c d e f g h (PR: p a b c d e f g h : Prop), q a b c d e f g h : Prop) 
  (at level 50, no associativity).

Notation "p <9= q" := 
  (forall a b c d e f g h i (PR: p a b c d e f g h i : Prop), q a b c d e f g h i : Prop) 
  (at level 50, no associativity).

Notation "p <10= q" := 
  (forall a b c d e f g h i j (PR: p a b c d e f g h i j : Prop), q a b c d e f g h i j : Prop) 
  (at level 50, no associativity).

Notation "p <11= q" := 
  (forall a b c d e f g h i j k (PR: p a b c d e f g h i j k : Prop), q a b c d e f g h i j k : Prop) 
  (at level 50, no associativity).

Notation "p <12= q" := 
  (forall a b c d e f g h i j k l (PR: p a b c d e f g h i j k l : Prop), q a b c d e f g h i j k l : Prop) 
  (at level 50, no associativity).

Notation "p \1/ q" := 
  (fun a => p a \/ q a) 
  (at level 50, no associativity).

Notation "p \2/ q" := 
  (fun a b => p a b \/ q a b) 
  (at level 50, no associativity).

Notation "p \3/ q" := 
  (fun a b c => p a b c \/ q a b c) 
  (at level 50, no associativity).

Notation "p \4/ q" := 
  (fun a b c d => p a b c d \/ q a b c d) 
  (at level 50, no associativity).

Notation "p \5/ q" := 
  (fun a b c d e => p a b c d e \/ q a b c d e) 
  (at level 50, no associativity).

Notation "p \6/ q" := 
  (fun a b c d e f => p a b c d e f \/ q a b c d e f) 
  (at level 50, no associativity).

Notation "p \7/ q" := 
  (fun a b c d e f g => p a b c d e f g \/ q a b c d e f g) 
  (at level 50, no associativity).

Notation "p \8/ q" := 
  (fun a b c d e f g h => p a b c d e f g h \/ q a b c d e f g h) 
  (at level 50, no associativity).

Notation "p \9/ q" := 
  (fun a b c d e f g h i => p a b c d e f g h i \/ q a b c d e f g h i) 
  (at level 50, no associativity).

Notation "p \10/ q" := 
  (fun a b c d e f g h i j => p a b c d e f g h i j \/ q a b c d e f g h i j) 
  (at level 50, no associativity).

Notation "p \11/ q" := 
  (fun a b c d e f g h i j k => p a b c d e f g h i j k \/ q a b c d e f g h i j k) 
  (at level 50, no associativity).

Notation "p \12/ q" := 
  (fun a b c d e f g h i j k l => p a b c d e f g h i j k l \/ q a b c d e f g h i j k l) 
  (at level 50, no associativity).

Notation "p /1\ q" := 
  (fun a => p a /\ q a) 
  (at level 50, no associativity).

Notation "p /2\ q" := 
  (fun a b => p a b /\ q a b) 
  (at level 50, no associativity).

Notation "p /3\ q" := 
  (fun a b c => p a b c /\ q a b c) 
  (at level 50, no associativity).

Notation "p /4\ q" := 
  (fun a b c d => p a b c d /\ q a b c d) 
  (at level 50, no associativity).

Notation "p /5\ q" := 
  (fun a b c d e => p a b c d e /\ q a b c d e) 
  (at level 50, no associativity).

Notation "p /6\ q" := 
  (fun a b c d e f => p a b c d e f /\ q a b c d e f) 
  (at level 50, no associativity).

Notation "p /7\ q" := 
  (fun a b c d e f g => p a b c d e f g /\ q a b c d e f g) 
  (at level 50, no associativity).

Notation "p /8\ q" := 
  (fun a b c d e f g h => p a b c d e f g h /\ q a b c d e f g h) 
  (at level 50, no associativity).

Notation "p /9\ q" := 
  (fun a b c d e f g h i => p a b c d e f g h i /\ q a b c d e f g h i) 
  (at level 50, no associativity).

Notation "p /10\ q" := 
  (fun a b c d e f g h i j => p a b c d e f g h i j /\ q a b c d e f g h i j) 
  (at level 50, no associativity).

Notation "p /11\ q" := 
  (fun a b c d e f g h i j k => p a b c d e f g h i j k /\ q a b c d e f g h i j k) 
  (at level 50, no associativity).

Notation "p /12\ q" := 
  (fun a b c d e f g h i j k l => p a b c d e f g h i j k l /\ q a b c d e f g h i j k l) 
  (at level 50, no associativity).

Ltac esplits :=
  intros;
  repeat match goal with 
  | [ |- @ex _ _ ] => eexists
  | [ |- _ /\ _ ] => split
  end.

Ltac exact_auto H := first
  [ first [exact H | eexact H] 
  | exact_auto (proj1 H)
  | exact_auto (proj2 H) ].
    
Ltac exactH :=
  try (match goal with [ H : _ |- _ ] => exact_auto H end; eauto).

Ltac applyH :=
  try (match goal with [ H : _ |- _ ] => first [apply H | eapply H] end; eauto).

Ltac rewrite_auto H := first
  [ rewrite H
  | rewrite_auto (proj1 H)
  | rewrite_auto (proj2 H) ].
Ltac rewrite_auto_r H := first 
  [ rewrite <- H
  | rewrite_auto_r (proj1 H)
  | rewrite_auto_r (proj2 H) ].
Ltac rewrite_auto_h H H' := first
  [ rewrite H in H'
  | rewrite_auto_h (proj1 H) H'
  | rewrite_auto_h (proj2 H) H'].
Ltac rewrite_auto_h_r H H' := first 
  [ rewrite <- H in H'
  | rewrite_auto_h_r (proj1 H) H'
  | rewrite_auto_h_r (proj2 H) H'].

Tactic Notation "rewriteD" constr(H) := (rewrite_auto H).
Tactic Notation "rewriteD" "<-" constr(H) := (rewrite_auto_r H).
Tactic Notation "rewriteD" constr(H) "in" hyp(H') := (rewrite_auto_h H H').
Tactic Notation "rewriteD" "<-" constr(H) "in" hyp(H') := (rewrite_auto_h_r H H').

Tactic Notation "rewriteH" := 
  match goal with [ H : _ |- _ ] => rewriteD H end.
Tactic Notation "rewriteH" "<-" := 
  match goal with [ H : _ |- _ ] => rewriteD <- H end.

Tactic Notation "rewriteH" "in" hyp(H') := 
  match goal with [ H : _ |- _ ] => rewriteD H in H' end.
Tactic Notation "rewriteH" "<-" "in" hyp(H') := 
  match goal with [ H : _ |- _ ] => rewriteD <- H in H' end.

Ltac mono H tac := let x := fresh "x" in let H1 := fresh H in let H2 := fresh H in 
  first[
    (intro H1; specialize (H H1); mono H tac)
  | (destruct H as (x & H); exists x; mono H tac)
  | (destruct H as (H1 & H2); split; [mono H1 tac| mono H2 tac]) 
  | (destruct H as [H1 | H2]; [left; mono H1 tac| right; mono H2 tac]) 
  | by applyH
  | tac H
  ].

Definition option_default {A: Type} {B: Type} (b: B) (f: A -> B) (oa: option A) : B :=
  match oa with
  | None => b
  | Some a => f a
  end.

Program Definition option_drop {A} (oa: option A) (Hsome: oa <> None) : A :=
  match oa with
  | Some a => a
  | None => _
  end.

Lemma option_drop_eq: forall {A} (oa: option A) (Hsome: oa <> None),
  oa = Some (option_drop oa Hsome).
Proof.
intros; destruct oa; auto.
by elimtype False; apply Hsome.
Qed.

Lemma option_dec: forall {A} (a: option A), {a = None} + {a <> None}.
Proof. by intros; destruct a; [right; discriminate|left]. Qed.

Lemma option_eq: forall {A} (a b: option A),
  (forall c, a = Some c <-> b = Some c) <-> a = b.
Proof. 
split; intros.
  by destruct a; [symmetry; applyH | destruct b; try applyH].
by subst; intuition.
Qed.

Lemma mp: forall P Q: Type, P -> (P -> Q) -> Q.
Proof. intuition. Defined.

Lemma and_sym: forall P Q, P /\ Q -> Q /\ P.
Proof. by intuition. Qed.
