Set Implicit Arguments.

Open Scope type_scope.

(*************************
   Stream 
 *************************)

CoInductive pstream A (R: Type) : Type :=
| scons (a: A) (s: pstream A R + R)
.

Definition itrstrcore A L R (f: L -> pstream A (L + R)) 
                     (itrstr': pstream A (L + R) -> pstream A R)
                     (ts: pstream A (L + R) + (L + R)) : pstream A R + R :=
  match ts with
  | inl t => inl (itrstr' t)
  | inr (inl t) => inl (itrstr' (f t))
  | inr (inr t) => inr t
  end.

CoFixpoint itrstr' A L R (f: L -> pstream A (L + R)) 
                 (s: pstream A (L + R)) : pstream A R :=
  match s with
  | scons a s' => scons a (itrstrcore f (itrstr' f) s')
  end.

Definition itrstr A L R (f: L -> pstream A (L + R)) (s : L) : pstream A R :=
  itrstr' f (f s).

Definition hd A (s: pstream A False) : A :=
  match s with
  | scons a s' => a
  end.

Definition tl A (s: pstream A False) : pstream A False :=
  match s with
  | scons a (inl s') => s'
  | scons a (inr x) => False_rect _ x
  end.

(* ones := [1,1,1,...] *)
Definition ones : pstream nat False :=
  itrstr (fun ones => scons 1 (inr (inl ones))) tt.

Eval compute in (hd ones).
Eval compute in (hd (tl ones)).
Eval compute in (hd (tl (tl ones))).
Eval compute in (hd (tl (tl (tl ones)))).

(* flip := [1,2,1,2,...] *)
Definition flip : pstream nat False :=
  itrstr (fun flip => scons 1 (inl (scons 2 (inr (inl flip))))) tt.

Eval compute in (hd flip).
Eval compute in (hd (tl flip)).
Eval compute in (hd (tl (tl flip))).
Eval compute in (hd (tl (tl (tl flip)))).

(*************************
   Infinite Binary Tree
 *************************)

CoInductive cotree A R : Type :=
| tcons (a: A) (l: cotree A R + R) (r: cotree A R + R)
.

Definition itrcotcore A L R (f: L -> cotree A (L + R)) 
                     (itrcot': cotree A (L + R) -> cotree A R)
                     (ts: cotree A (L + R) + (L + R)) : cotree A R + R :=
  match ts with
  | inl t => inl (itrcot' t)
  | inr (inl t) => inl (itrcot' (f t))
  | inr (inr t) => inr t
  end.

CoFixpoint itrcot' A L R (f: L -> cotree A (L + R)) 
                 (t: cotree A (L + R)) : cotree A R :=
  match t with
  | tcons a t1 t2 => tcons a (itrcotcore f (itrcot' f) t1) (itrcotcore f (itrcot' f) t2)
  end.

Definition itrcot A L R (f: L -> cotree A (L + R)) (s : L) : cotree A R :=
  itrcot' f (f s).

Definition root A (t: cotree A False) : A :=
  match t with
  | tcons a _ _ => a
  end.

Definition tleft A (t: cotree A False) : cotree A False :=
  match t with
  | tcons a (inl s') _ => s'
  | tcons a (inr x) _ => False_rect _ x
  end.

Definition tright A (t: cotree A False) : cotree A False :=
  match t with
  | tcons a _  (inl s') => s'
  | tcons a _ (inr x) => False_rect _ x
  end.

(* mytree :=

      1 (= x )
     / \ 
    x   2 (= y )
      /  \ 
     x    y
*)

Definition fliptree : cotree nat False :=
  itrcot (fun onetree => tcons 1 
    (inr (inl onetree)) 
    (inl (itrcot (fun twotree => tcons 2
            (inr (inr (inl onetree)))
            (inr (inl twotree))) I))) I.

Eval compute in (root fliptree).
Eval compute in (root (tleft fliptree)).
Eval compute in (root (tleft (tleft fliptree))).
Eval compute in (root (tright (tleft fliptree))).
Eval compute in (root (tright fliptree)).
Eval compute in (root (tleft (tright fliptree))).
Eval compute in (root (tright (tright fliptree))).
Eval compute in (root (tleft (tright (tright fliptree)))).
Eval compute in (root (tright (tright (tright fliptree)))).

