Require Export cola.
Require Export paconotation.
Require Export pacotac.

Open Scope cola_scope.
Set Implicit Arguments.

(* 
   Parameterized Coinduction  
*)

Section parameterized_coinduction.

Variable C : cola.
Variable gf : C -> C.

(* upper: mendler style monotonization function *)

Definition upper (x: C) := cola_lub (fun y => exists z, z <= x /\ y <= gf z).

Lemma upper_mon: monotone upper.
Proof.
  repeat intro; apply cola_lub_mon; repeat intro.
  destruct IN as (?,(?,?)); exists x1; split; [rewrite H|]; eauto.
Qed.

(* Thanks to the use of upper in the definition of G,
   we can remove monotonicity condition from 
   G_acc, G_mon, G_mult and G_fold.
*)

Definition G r := cofixp (fun x => upper (x \\// r)).

Theorem G_acc: forall l r0 
    (OBG: forall r (INC: r0 <= r) (CIH: l <= r), l <= G r),
  l <= G r0.
Proof. 
  unfold monotone in *; repeat intro.
  eapply cola_trans with (y:= G (r0 \\// l)). 
    eauto using cola_join_upper_l, cola_join_upper_r.
  unfold G in *; eapply cofixp_coind. 
  rewrite cofixp_le at 1.
  2: repeat intro; apply upper_mon; eauto using cola_join_mon, cola_join_upper_l. 
  eapply cola_lub_least; intros; eapply cola_lub_upper.
  destruct H as (?,(?,?)).
  exists x; split; eauto; rewrite H.
  eauto using cola_trans, cola_join_least, cola_join_upper_l, cola_join_upper_r.
Qed.

Theorem G_mon: monotone G.
Proof.
  repeat intro; eapply cofixp_mon. 
  repeat intro; eapply upper_mon; eauto using cola_join_mon. 
Qed.

Theorem G_mult_strong: forall r, G (G r \\// r) <= G r.
Proof. 
  intros; apply cofixp_coind.
  eapply cola_trans; [eapply cofixp_le; repeat intro; apply upper_mon|]; 
   eauto using upper_mon, G_mon, cola_join_mon, 
               cola_join_least, cola_join_upper_l, cola_join_upper_r.
Qed.

Corollary G_mult: forall r, G (G r) <= G r.
Proof.
  intros; eapply cola_trans, G_mult_strong; eapply G_mon, cola_join_upper_l.
Qed.

Theorem G_fold: forall r, gf (G r \\// r) <= G r.
Proof. 
  intros; eapply cola_trans; [|apply le_cofixp].
  2: repeat intro; apply upper_mon; eauto using cola_join_mon, cola_join_upper_l. 
  eapply cola_lub_upper; eauto.
Qed.

Theorem G_unfold: forall (MON: monotone gf) r, G r <= gf (G r \\// r).
Proof. 
  intros; eapply cola_trans; [apply cofixp_le|].
    repeat intro; apply upper_mon; eauto using cola_join_mon, cola_join_upper_l. 
  eapply cola_lub_least; intros; destruct H as (?,(?,?)); rewrite H0; eauto.
Qed.

End parameterized_coinduction.

Implicit Arguments G [C].
Arguments G [C] gf r : simpl never.

Implicit Arguments G_acc [C].

Implicit Arguments G_mon [C].

Implicit Arguments G_mult_strong [C].

Implicit Arguments G_mult [C].

Implicit Arguments G_fold [C].
Hint Resolve G_fold.

Implicit Arguments G_unfold [C].
