Require Import String List.
Require Import vgtac basis lang.

Set Implicit Arguments.

(**********************************
   Definitions about reduces
 **********************************)

Inductive taureduces e : exp -> Prop :=
| taur_nil: taureduces e e
| taur_step: forall em e' 
    (STEP: reduce e Qtau em)
    (RED: taureduces em e'),
  taureduces e e'.
Hint Resolve taur_step.

Inductive qreduce (e: exp): evt -> exp -> Prop :=
| qred_nil: qreduce e Qtau e
| qred_step: forall q e' (RED: reduce e q e'), qreduce e q e'.
Hint Resolve qred_step.

CoInductive Obs :=
| Onil
| Oevt (oq: oevt) (oqs: Obs)
.

CoInductive produce_obs (e : exp) : Obs -> Prop :=
| prd_nil : produce_obs e Onil
| prd_evt : forall e' e'' (oq: oevt) oqs
    (TAU: taureduces e e') (RED: reduce e' oq e'') (PRD: produce_obs e'' oqs),
  produce_obs e (Oevt oq oqs)
.

(**********************************
   Lemmas
 **********************************)

Lemma taureduces_trans: forall e e' e'' 
  (R: taureduces e e') (R': taureduces e' e''), taureduces e e''.
Proof. induction 1; eauto. Qed.

Lemma taureduces_scont: forall (ke : skes) (e' : exp),
    taureduces (body_of ke) e' ->
  taureduces (term_of ke) (appctx_scont_exp (scont_of ke) e').
Proof. i; simplcanon; revert sk; induction H; eauto using taur_nil. Qed.

Lemma qreduce_scont: forall (ke : skes) (q : evt) (e' : exp),
    qreduce (body_of ke) q e' ->
  qreduce (term_of ke) q (appctx_scont_exp (scont_of ke) e').
Proof. i; simplcanon; destruct H; s; [econs 1|econs 2]; eauto. Qed.

Lemma val_nored :
   forall (ve: voe) q e', ~ reduce (e_of ve) q e'.
Proof.
  i; remember (e_of ve) as e; revert ve Heqe.
  induction FH; try by intros []; i; subst; destruct v_of.
  i; simplcanon; cut (exists v : val, e = v); [by intro; des; eauto|].
  revert v Heqe; clear; i; simplcanon; destruct SK, v; inv Heqe; eauto.
Qed.

Lemma reduce_inv_scont: forall (ke: skes) q e
    (RED: reduce (term_of ke) q e)
    (NOTVAL: forall ve: voe, ~ e_of ve = body_of ke ),
  exists e', reduce (body_of ke) q e' /\ e = appctx_scont_exp (scont_of ke) e'.
Proof.
  i; simplcanon; depdes RED; try by destruct sk; inv x; exfalso; eauto.
  simplcanon; assert (X: SK = sk /\ e = e0).
    by destruct SK, sk; depdes x; simplcanon; eauto; 
       exfalso; eauto using val_nored.
  des; subst; eauto.
Qed.

Lemma reduce_det: forall e0 q e q' e' 
    (RED: reduce e0 q e) 
    (RED': reduce e0 q' e'),
  match q with
    | OQin n => (q' = q -> e' = e) /\ (exists m, q' = OQin m)
    | _ => q' = q /\ e' = e
  end.
Proof.
  i; revert q' e' RED'; induction RED; i; subst;
    try by depdes RED'; simplcanon; esplits; i; eauto; try inv H;
           destruct SK; depdes x; exfalso; eauto using val_nored.
  eapply reduce_inv_scont in RED'.
  - des; subst; eapply IHRED in RED'.
    destruct q as [|[]]; des; subst; eauto.
    esplits; eauto; i; inv H; f_equal; eauto.
  - i; simplcanon; eauto using val_nored.
Grab Existential Variables. exact 0.
Qed.

Definition beq_input e : bool := match e with | Einput => true | _ => false end.

Lemma reduce_inv: forall ke q e q' e'
    (RED: reduce (redex_of ke) q e) 
    (RED': reduce (exp_of ke) q' e'),
  match beq_input (redex_of ke) with
  | true => exists m, q' = OQin m /\ e' = app_cont (cont_of ke) (En m)
  | false => match q with
            | OQin _ => True
            | _ => q' = q /\ e' = app_cont (cont_of ke) e
           end
  end.
Proof.
  i; simplcanon; clear notval; destruct e0;
    try by eapply reduce_det in RED'; [|by clear RED'; eauto];
           destruct q as [|[]]; des; subst; eauto.
  s; clear q e RED; assert (X:=RED'); eapply reduce_det in X; [|by clear X RED'; eauto].
  ss; des; subst; eapply reduce_det in RED'; [|by clear RED'; eauto].
  ss; des; subst; eauto. 
Grab Existential Variables. exact 0.
Qed.

Lemma subst_exp_disj : 
  forall (x: exp) sub (H: list_disj eq_termvar (map (@fst _ _) sub) (fv_exp x)), subst_exp sub x = x.
Proof. induction x; subst_tac. Qed. 

Lemma subst_exp_closed : 
  forall (x: exp) (H: fv_exp x = nil) sub, subst_exp sub x = x.
Proof. intros; apply subst_exp_disj; rewrite H; constructor. Qed.

Lemma value_dec: forall e, 
  (exists ve:voe, e_of ve = e) \/ (~ (exists ve:voe, e_of ve = e)).
Proof.
  i; induction e; 
    by des; subst; eauto; right; 
       i; des; simplcanon; destruct v; inv FH; simplcanon; eauto.
Qed.
