Require Import String Arith Program.
Require Import basis vgtac lang langlemmas.

Set Implicit Arguments.


(******************************
   Simulations and similarity
 ******************************)

Definition exprel := exp -> exp -> Prop.

Inductive sim (r: exprel) (e1 e2: exp) : Prop :=
| _simul
  (VAL: forall (ve: voe) (VAL: e_of ve = e1), taureduces e2 (e_of ve))
  (EXP: forall q e1' (RED: reduce e1 q e1'),
        exists e2m e2', 
          <<TAU: taureduces e2 e2m>> /\ 
          <<RED: qreduce e2m q e2'>> /\
          <<REL: r e1' e2'>>).

(********************************
   Tactics for simulation
 ********************************)

Ltac simul_inv := let RED := fresh "RED" in let VAL := fresh "VAL" in
  i; econs; [by i; exfalso; eapply notval, VAL|]; 
  i; eapply reduce_inv in RED; [|first [by apply (Rinput 0)|by econs]]; 
  inst; s in RED; des.

Tactic Notation "simul_step" int_or_var(n) :=
  simul_inv; match goal with 
  [H:_=Qtau|-_] => subst; esplits; 
     [ do n (econs 2; [step|inst;s]); econs; reflexivity | econs; reflexivity | inst; s]
  | _  => subst; esplits; 
     [ repeat (econs 2; [step|inst;s]); econs; reflexivity | econs 2; step | inst; s]
  end. 

Hint Extern 3 (@eq nat _ _) => omega.

(********************************
   Example: Three recursive programs
 ********************************)

Definition restart : exp :=
  Efix "pr" "f" (Efix "_" "n" (
    If "n" ~<= 0 then "f" @ 0 else Eoutput "n" ;; "pr" @ "f" @ ("n" ~- 1)
  )).

Definition f : exp :=
  Efix "f" "n" (
    Let "v" := (Eoutput ("n") ;; Einput) ~* 2 in 
    (If 1 ~<= "v" then "f" else restart @ "f") @ ("v" ~+ "n")).

Definition g : exp :=
  Efix "g" "m" (
    Eoutput (2 ~* "m") ;;
    Let "v" := Einput in 
    If "v" ~<= 0 then restart @ "g" @ (2 ~* "m") else "g" @ ("v" ~+ "m")).

Ltac fold_rfg := fold restart f g.
Ltac unfold_rfg := unfold f, g, restart.
