Require Import String Arith Program.
Require Import basis vgtac lang langlemmas.
Require Import paco.
Require Import simul_def.

Set Implicit Arguments.

(**********************************************
   Verification using parameterized coinduction (internal approach)
 **********************************************)

Tactic Notation "psimil_step" int_or_var(n) := 
  try left; pfold; simul_step n.

Lemma rfg_simulated: forall n m (EQ: n = 2 * m), 
  paco2 sim bot2 (f @ n) (g @ m).
Proof.
  pcofix CIH.
  intros; subst; do 6 psimil_step 1; do 2 psimil_step 0.
  destruct m0; psimil_step 2; [|by eauto].
  left; fold_rfg; generalize (m+(m+0)).
  pcofix CIH'.
  intros; do 3 psimil_step 1.
  destruct n; psimil_step 1; [by eauto|].
  do 3 psimil_step 1; eauto.
Qed.

(**********************************************
   Compositional verification
 **********************************************)

Lemma restart_simulated: forall (f g: voe) (n m: nat) (r: exprel) 
    (ASM: r ((e_of f) @ 0) ((e_of g) @ 0))
    (CLOf: fv_exp (e_of f) = []) (CLOg: fv_exp (e_of g) = []) (EQ: n = m),
  paco2 sim r (restart @ (e_of f) @ n) (restart @ (e_of g) @ m).
Proof.
  pcofix CIH; intros; simplcanon.
  do 3 psimil_step 1.
  destruct ble_nat; psimil_step 1.
  - by right; rewrite !subst_exp_closed; eauto.
  - do 3 psimil_step 1.
    right; rewrite !subst_exp_closed; eauto.
Qed.

Lemma fg_simulated_using_lemma: forall n m (EQ: n = 2 * m),
  paco2 sim bot2 (f @ n) (g @ m).
Proof.
  pcofix CIH; i; subst.
  do 6 psimil_step 1; do 2 psimil_step 0.
  destruct m0; psimil_step 2; [|by eauto]. 
  left; eapply restart_simulated; s; eauto.
Qed.
